import asyncio
import functools
from aiossdb.pool import create_pool


class Client:
    def __init__(self, host='127.0.0.1', port=8888, password=None, timeout=None, max_connection=100, loop=None):
        self.host = host
        self.port = port
        self.password = password
        self.timeout = timeout
        self.max_connection = max_connection

        if loop is None:
            loop = asyncio.new_event_loop()
            asyncio.set_event_loop(loop)
        self.loop = loop

        self._pool = None

    async def get_pool(self):
        if self._pool is None:
            self._pool = await create_pool(
                (self.host, self.port), password=self.password, loop=self.loop,
                timeout=self.timeout, maxsize=self.max_connection
            )
        return self._pool

    async def execute(self, cmd, *args):
        pool = await self.get_pool()
        res = await pool.execute(cmd, *args)
        return res

    def __getattr__(self, item):
        if item not in self.__dict__:
            self.__dict__[item] = functools.partial(self.execute, item)

        return self.__dict__[item]

    async def close(self):
        if self._pool:
            self._pool.close()
            await self._pool.wait_closed()
            self._pool = None



import asyncio
import collections

from .connection import create_connection
from .errors import PoolClosedError
from .log import logger


async def create_pool(address, *, password=None, encoding='utf-8', minsize=1, maxsize=10,
                parser=None, loop=None, timeout=None, pool_cls=None, connection_cls=None):
    if pool_cls is None:
        pool_cls = SSDBConnectionPool

    pool = pool_cls(address, password=password, encoding=encoding,
                    parser=parser, minsize=minsize, maxsize=maxsize,
                    loop=loop, timeout=timeout, connection_cls=connection_cls)

    # 首先先填充空闲连接
    try:
        # 填充至最小minsize的大小
        await pool._fill_free(overall=False)
    except Exception as e:
        pool.close()
        await pool.wait_closed()
        raise
    return pool


class SSDBConnectionPool:

    def __init__(self, address, *, password=None, parser=None, encoding=None, minsize, maxsize,
                 connection_cls=None, timeout=None, loop=None):
        assert isinstance(minsize, int) and minsize >= 0, ("minsize must be int >=0", minsize, type(minsize))
        assert isinstance(maxsize, int) and maxsize >= minsize, (
            "maxsize must be int >= minsize", maxsize, type(maxsize), minsize)
        if loop is None:
            loop = asyncio.new_event_loop()
            asyncio.set_event_loop(loop)
        self._address = address
        self._password = password
        self._parser_class = parser
        self._timeout = timeout
        self._loop = loop
        self._used = set()
        self._connection_cls = connection_cls
        self._pool = collections.deque(maxlen=maxsize)
        self._minsize = minsize
        self._maxsize = maxsize
        self._encoding = encoding
        # 用于release后同步各个其他获取新连接的协程，使其开始工作，否则等待条件
        self._cond = asyncio.Condition(lock=asyncio.Lock(loop=loop), loop=loop)
        self._waiter = None
        self._closing = False
        self._closed = False

    async def execute(self, command, *args, **kwargs):
        conn, address = await self.get_connection()
        try:
            fut = await conn.execute(command, *args, **kwargs)
        finally:
            await self.release(conn)
        return fut

    @property
    def minsize(self):
        return self._minsize

    @property
    def maxsize(self):
        return self._maxsize

    @property
    def freesize(self):
        return len(self._pool)

    @property
    def size(self):
        return len(self._pool) + len(self._used)

    async def get_connection(self):
        """获取连接，要么在空闲连接中直接获取，要么等待直到获得新的连接，
        不论如何获取的连接，都要进入used连接集合中, 然后使用完成之后返回pool可用连接中

        如果连接池自己调用的execute会自动调用release
        如果直接调用这个函数获取的连接，使用完成之后必须显式调用release方法"""
        # 在pool中寻找
        for i in range(self.freesize):
            conn = self._pool.popleft()
            # 如果连接已经关闭，则查找pool中下一个连接
            if conn.closed:
                continue
            self._used.add(conn)
            return conn, conn.address
        # 如果pool中已经没有可用连接了，动态获取连接
        conn = await self.new_connection()
        return conn, conn.address

    async def new_connection(self):
        """pool中无可用连接，在这里创建新连接填充pool
        有可能本身used的size已经是maxsize，则填充失败，需要监听release方法的信号通知
        来重新填充，最后返回一条连接"""
        if self.closed:
            raise PoolClosedError("Pool is closed")
        with (await self._cond):
            # 获得锁之后可能pool已经关闭，所以进行两遍检测
            if self.closed:
                raise PoolClosedError("Pool is closed")
            # 下面获取首先填充pool，然后获取新连接
            # 有可能连接池已经满了，无法获取新连接
            while 1:
                # fill_free创建新连接,直接将pool的可用额度填满
                # 有种情况就是self.size已经满了，但是都是在used，所以就得等待release
                await self._fill_free(overall=True)
                if self.freesize:
                    conn = self._pool.popleft()
                    # 排除一些错误情况
                    if conn.closed or conn in self._used:
                        logger.error("conn is closed or conn is in used, {} - {}".format(conn, self._used))
                        continue
                    self._used.add(conn)
                    return conn
                else:
                    # 等待release的释放连接，然后调用notify方法来通知此处
                    # wait的时候会将lock释放
                    await self._cond.wait()

    async def _fill_free(self, *, overall):
        """填充pool连接池,应该填充使得有可用连接，或者填充到self._minsize"""
        self._drop_closed()
        # 首先将size填充到最小连接数
        while self.size < self._minsize:
            try:
                conn = await create_connection(
                    self._address, password=self._password, encoding=self._encoding, parser=self._parser_class,
                    loop=self._loop, timeout=self._timeout, connect_cls=self._connection_cls
                )
            except Exception as e:
                logger.error("create connection encountered error: {}".format(e))
            else:
                self._pool.append(conn)
        if self.freesize:
            return
        if overall:
            # 一直填充到可用连接池中有连接，并且size应该小于最大size
            while not self._pool and self.size < self.maxsize:
                try:
                    conn = await create_connection(
                        self._address, password=self._password, encoding=self._encoding, parser=self._parser_class,
                        loop=self._loop, timeout=self._timeout, connect_cls=self._connection_cls
                    )
                except Exception as e:
                    logger.error("create connection encountered error: {}".format(e))
                else:
                    self._pool.append(conn)

    async def release(self, conn):
        """将没有关闭的连接从used集合放回可用的pool中，或者关闭仍然有命令的连接
        并且给一个信号通知，这样获取新连接的方法就可以获取新连接"""
        # 关闭的时候已经清空pool和used了
        if self.closed:
            raise PoolClosedError("Pool is closed")
        assert conn in self._used, ("Invalid connection, maybe from other pool", conn)
        self._used.remove(conn)
        # 如果连接还未关闭，则置入可用连接池
        if not conn.closed:
            self._pool.append(conn)
        else:
            # 如果已经关闭，则不管理
            logger.warn("Connection {} has been closed".format(conn))

        # 在这里提供信号量通知
        asyncio.ensure_future(self._wake_up(), loop=self._loop)

    async def _wake_up(self):
        with (await self._cond):
            # 通知其他协程开始工作
            self._cond.notify()

    def _drop_closed(self):
        """清除关闭的连接，pool里的和used的"""
        for i in range(self.freesize):
            conn = self._pool[0]
            if conn.closed:
                self._pool.popleft()
            else:
                # 将其转到队列尾端
                self._pool.rotate(1)
        # 使用池中已经关闭的连接
        _closed_used = set()
        for conn in self._used:
            if conn.closed:
                _closed_used.add(conn)
        # 两者求差集
        self._used = self._used.difference(_closed_used)

    def close(self):
        """关闭所有的连接，pool以及正在使用的连接"""
        self._closing = True
        self._waiter = asyncio.ensure_future(self._do_close(), loop=self._loop)

    async def _do_close(self):
        """将所有的pool连接和used连接取出来，可能需要等待，加入期物列表"""
        with (await self._cond):
            waiters = []
            while self._pool:
                conn = self._pool.popleft()
                conn.close()
                # 加入期物
                waiters.append(conn.wait_closed())
            for conn in self._used:
                conn.close()
                waiters.append(conn.wait_closed())
            await asyncio.gather(*waiters, loop=self._loop)
            self._closed = True

    async def wait_closed(self):
        """等待直到连接池关闭，这里要等待的是所有连接池连接的关闭期物Future的完成"""
        await asyncio.shield(self._waiter, loop=self._loop)

    @property
    def closed(self):
        return self._closing or self._closed

    async def auth(self, password):
        """将pool里面的每个连接进行auth"""
        self._password = password
        with (await self._cond):
            for i in range(self.freesize):
                await self._pool[i].auth(password)

    def __repr__(self):
        return '<{} [size:[{}:{}], free:{}]>'.format(
            self.__class__.__name__, self._minsize, self._maxsize, self.freesize
        )



from .errors import *


def utf8_encode(s):
    """只支持Python3"""
    s = str(s) if isinstance(s, int) else s
    return s.encode('utf8') if isinstance(s, str) else s


def encode_command(command, *args):
    """将命令转换成协议要求的命令格式
    Request := Cmd Blocks*
    Cmd     := Block
    Block  := Size '\n' Data '\n'
    Size   := literal_integer
    Data   := size_bytes_of_data
    比如
        3
        get
        3
        key
    """
    if command == "delete":
        command = "del"

    args = [utf8_encode(command)] + [utf8_encode(i) for i in args]
    buf = utf8_encode('').join(utf8_encode('%d\n%s\n') % (len(i), i) for i in args) + utf8_encode('\n')
    return buf


class SSDBParser:
    def __init__(self, encoding=None):
        # 字节数组
        self.buf = bytearray()
        self.pos = 0
        self._gen = None
        self.encoding = encoding

    def feed(self, data, o=0, l=-1):
        if l == -1:
            l = len(data) - o
        if o < 0 or l < 0:
            raise ValueError("negative input")
        if o + l > len(data):
            raise ValueError("input is larger than buffer size")
        self.buf.extend(data[o:o+l])

    def gets(self):
        """获取解析的数据，或者返回False"""
        return self.parse_one()

    def wait_some(self, size):
        """如果buf长度一直小于所需要的size，则一直迭代False，
        这样就要从套接字来读取数据，直到拿到所有数据"""
        while len(self.buf) < self.pos + size:
            yield False

    def wait_any(self):
        """等待buf进入任意一个字符"""
        yield from self.wait_some(len(self.buf) + 1)

    def read_line(self, size=None):
        if size is not None:
            # size是要读取的数据长度，不会包含\n
            if len(self.buf) < size + 1 + self.pos:
                yield from self.wait_some(size+1)
            offset = self.pos + size
            # 如果请求的该行结束符不是\n的话
            if self.buf[offset:offset+1] != b'\n':
                raise ProtocolError(msg="Expected b'\n'")
        else:
            # 从self.pos开始查找b'\n'的最早出现的位置，如果没有返回-1
            offset = self.buf.find(b'\n', self.pos)
            while offset < 0:
                yield from self.wait_any()
                offset = self.buf.find(b'\n', self.pos)
        val = self.buf[self.pos:offset]
        # 读取一行后重置
        self.pos = 0
        # 删除该行数据
        del self.buf[:offset + 1]
        if self.encoding:
            val = val.decode(self.encoding)
        return val

    def read_int(self):
        """读取协议中定义为数据长度的行"""
        try:
            value = yield from self.read_line()
            return int(value)
        except ValueError:
            raise ProtocolError("Expected int")

    def parse(self):
        size = yield from self.read_int()
        status = yield from self.read_line(size)
        if status != 'ok':
            return ReplyError(status)
        data = []
        try:
            # 可能没有数据，所以在读完状态后的值可能不是int，而是换行符
            # 如果有异常则数据为空
            size = yield from self.read_int()
        except ProtocolError:
            return data
        while True:
            val = yield from self.read_line(size)
            data.append(val)
            try:
                size = yield from self.read_int()
            except ProtocolError:
                break
        return data

    def parse_one(self):
        # 用来驱动生成器获取需要的数据
        # 读取一行，偶数行是数字，奇数行是数据
        if self._gen is None:
            self._gen = self.parse()
        try:
            # 激活协程
            self._gen.send(None)
        except StopIteration as e:
            self._gen = None
            return e.value
        except Exception:
            self._gen = None
            raise
        else:
            return False



# import os
# import sys
import logging


logger = logging.getLogger('aiossdb')

# if os.environ.get("AIOSSDB_DEBUG"):
#     logger.setLevel(logging.DEBUG)
#     handler = logging.StreamHandler(stream=sys.stderr)
#     handler.setFormatter(logging.Formatter(
#         "%(asctime)s %(name)s %(levelname)s %(message)s"))
#     logger.addHandler(handler)
#     os.environ["AIOSSDB_DEBUG"] = ""



from .connection import create_connection, SSDBConnection
from .errors import SSDBError, ReplyError, ConnectionClosedError, ProtocolError, PoolClosedError
from .parser import SSDBParser
from .pool import create_pool, SSDBConnectionPool
from .client import Client

__version__ = '0.0.2'
__author__ = 'Kevin Du'
__email__ = 'dgt_x@foxmail.com'
__licence__ = 'MIT'



class SSDBError(Exception):
    """aiossdb异常基类"""


class ConnectionClosedError(SSDBError):
    """如果到服务器的连接被关闭引发该异常"""


class ReplyError(SSDBError):
    """ssdb服务器返回的错误类型，可能有:
       not_found, error, fail, client_error
    """
    def __int__(self, etype, command=None):
        self.etype = etype
        self.command = command


class ProtocolError(SSDBError):
    """当解析协议出现问题时候引发该异常"""
    def __init__(self, msg):
        self.msg = msg


class PoolClosedError(SSDBError):
    """如果连接池已经关闭，引发该异常"""



import asyncio
import socket

from collections import deque

from .log import logger
from .parser import SSDBParser, encode_command
from .errors import ProtocolError, ReplyError, ConnectionClosedError
from .utils import wait_ok, set_result, set_exception


MAX_CHUNK_SIZE = 65536
_NOTSET = object()


async def create_connection(address, *, password=None, encoding='utf-8', parser=None, loop=None,
                      timeout=None, connect_cls=None, reusable=True):
    '''
    创建SSDB数据库连接
    :param address: 类似于socket的地址，如果是tuple或者list，则应该是(host, port)这种形式，
                    但是不支持unix socket
    :param password: SSDB数据库的密码，默认是None
    :param encoding: 用于将读取的数据从bytes解码成str，默认为None
    :param parser: 根据SSDB协议解析返回数据的解析器，默认会使用自定义的SSDBParser
    :param loop:
    :param timeout: 默认情况，timeout会在连接状态下应用限制等待时间，
                    也可以使用这个参数来定义创建连接所花的时间
    :param connect_cls:
    :param reusable: 设置端口重用，默认为True
    :return: 返回一个SSDBConnection对象，如果传递了connect_cls,则会返回这个类的实例
    '''
    # 首先判断address
    assert isinstance(address, (tuple, list)), "tuple or list expected"

    # 判断timeout
    if timeout is not None and timeout <= 0:
        raise ValueError("Timeout has to be None or a number greater than 0")

    # 默认connect_cls是SSDBConnection
    # TODO: 判断传入的类是否是可以应用的连接类型
    if connect_cls is None:
        connect_cls = SSDBConnection

    # 开始连接
    host, port = address
    logger.debug("Creating tcp connection to %r", address)
    # asyncio.open_connection创建套接字连接，返回reader和writer对象，它也是一个协程
    # 实际调用的是loop.create_connection
    # wait_for函数提供等待Future或者协程完成直到超时的功能，返回协程或者Future的结果
    reader, writer = await asyncio.wait_for(asyncio.open_connection(host, port, loop=loop), timeout, loop=loop)
    sock = writer.transport.get_extra_info('socket')
    if sock is not None:
        # 设置端口重用
        if reusable:
            sock.setsockopt(socket.SOL_SOCKET, socket.SO_REUSEADDR, 1)
        # 设置TCP无延迟，其相对是 Nagle’s Algorithm
        sock.setsockopt(socket.IPPROTO_TCP, socket.TCP_NODELAY, 1)
        address = sock.getpeername()
    address = tuple(address[:2])

    conn = connect_cls(reader, writer, encoding=encoding,
                       address=address, parser=parser, loop=loop)

    try:
        if password is not None:
            await conn.auth(password)
    except Exception:
        conn.close()
        await conn.wait_closed()
        raise
    return conn


class SSDBConnection:
    def __init__(self, reader, writer, *, address, encoding=None, parser=None, loop=None):
        if loop is None:
            # 默认使用asyncio的事件循环
            loop = asyncio.get_event_loop()
        if parser is None:
            parser = SSDBParser
        assert callable(parser), "Parser argument: {} is not callable".format(parser)
        self._reader = reader
        self._writer = writer
        self._address = address
        self._loop = loop
        # 使用双端队列来记录发送的命令，在解析数据的时候popleft
        self._waiters = deque()
        self._parser = parser(encoding=encoding)
        # 创建读取的task, self._read_data()是一个协程，用来在套接字生存期间读取数据
        # ensure_future 排定协程在事件循环的执行，如果参数是Future对象，将直接返回，返回的类型是Task对象
        self._reader_task = asyncio.ensure_future(self._read_data(), loop=self._loop)
        # 创建结束期物，这个任务会等待套接字读取任务的结束（使用回调来填充期物)
        self._close_waiter = asyncio.Future(loop=self._loop)
        # 添加读取任务结束后(套接字关闭)的回调函数
        self._reader_task.add_done_callback(self._close_waiter.set_result)
        self._encoding = encoding

        self._closing = False
        self._closed = False

    def __repr__(self):
        return '<SSDBConnection [host:{}-port:{}]>'.format(self._address[0], self._address[1])

    async def _read_data(self):
        # 在一个套接字生存期中，无限循环，直到断开连接，接收到EOF字符
        # self._reader.at_eof()在调用feed_eof()并且buffer为空的时候为True
        while not self._reader.at_eof():
            try:
                # 调用一个协程读取数据，每次只有全部读取完毕后才会返回数据
                data = await self._reader.read(MAX_CHUNK_SIZE)
            except asyncio.CancelledError:
                # 协程被取消，说明连接断开
                break
            except Exception as e:
                logger.error('Exception on data read %r', e, exc_info=True)
                break
            if data == b'' and self._reader.at_eof():
                # 如果读取数据为空，并且reader处于关闭状态，则说明服务器断开连接
                logger.debug('Connection has been closed by server')
            # 在这里解析器工作，解析数据
            self._parser.feed(data)
            # 获取数据,填充期物
            while 1:
                try:
                    obj = self._parser.gets()
                except ProtocolError as e:
                    self._do_close(e)
                    return
                else:
                    if obj is False:
                        break
                    # 这里将获取数据，填充期物（返回值）
                    self._process_data(obj)
        self._closing = True
        self._do_close(None)

    def _process_data(self, obj):
        assert len(self._waiters) > 0, (type(obj), obj)
        waiter, encoding, command = self._waiters.popleft()
        if isinstance(obj, ReplyError):
            obj.command = command
            set_exception(waiter, obj)
        else:
            set_result(waiter, obj)

    def execute(self, command, *args, encoding=_NOTSET):
        '''执行ssdb命令，返回期物等待结果'''
        if self._reader is None or self._reader.at_eof():
            raise ConnectionClosedError("Connection closed or corrupted")
        if command is None:
            raise TypeError("Command must not be None")
        if None in args:
            raise TypeError("args must not contain None")
        # 命令推荐小写
        command = command.lower().strip()

        if encoding is _NOTSET:
            encoding = self._encoding
        future = asyncio.Future(loop=self._loop)
        # 将命令和参数编码成协议要求的格式
        self._writer.write(encode_command(command, *args))
        # 将future进入队列，将来在接收到返回值的时候填充future
        self._waiters.append((future, encoding, command))
        return future

    def auth(self, password):
        future = self.execute('auth', password)
        return wait_ok(future)

    def close(self):
        self._do_close(None)

    def _do_close(self, exc):
        if self._closed:
            return
        self._closing = True
        self._closed = True
        self._writer.transport.close()
        self._reader_task.cancel()
        self._reader_task = None
        self._writer = None
        self._reader = None
        while self._waiters:
            # 将队列中还有的期物弹出并且取消
            waiter, *spam = self._waiters.popleft()
            logger.debug("Cancelling waiter %r", (waiter, spam))
            if exc is None:
                waiter.cancel()
            else:
                waiter.set_exception(exc)

    async def wait_closed(self):
        """协程 等待直到套接字连接关闭，防止self._close_waiter被取消"""
        await asyncio.shield(self._close_waiter, loop=self._loop)

    @property
    def closed(self):
        closed = self._closing or self._closed
        if not closed and self._reader and self._reader.at_eof():
            self._closing = closed = True
            # 立即调用该回调函数
            self._loop.call_soon(self._do_close, None)
        return closed

    @property
    def encoding(self):
        return self._encoding

    @property
    def address(self):
        return self._address



import asyncio
from .log import logger


def set_result(fut, result, *info):
    if fut.done():
        logger.debug("Waiter future is already done %r %r", fut, info)
        assert fut.cancelled(), (
            "waiting future is in wrong state", fut, result, info)
    else:
        fut.set_result(result)


def set_exception(fut, exception):
    if fut.done():
        logger.debug("Waiter future is already done %r", fut)
        assert fut.cancelled(), (
            "waiting future is in wrong state", fut, exception)
    else:
        fut.set_exception(exception)


async def wait_ok(fut):
    await fut
    return True




# coding=utf-8
import os
import re

from setuptools import setup


def get_version(package):
    init_py = open(os.path.join(package, '__init__.py')).read()
    return re.search("__version__ = ['\"]([^'\"]+)['\"]", init_py).group(1)


def get_packages(package):
    return [dirpath for dirpath, dirnames, filenames in os.walk(package)
            if os.path.exists(os.path.join(dirpath, '__init__.py'))]


with open('README.md', 'r', encoding='utf-8') as f:
    readme = f.read()

setup(
    name='aiossdb',
    version=get_version('aiossdb'),
    author='Kevin',
    author_email='dgt_x@foxmail.com',
    description='aiossdb is a library for accessing a ssdb database from the asyncio',
    long_description=readme,
    license='MIT',
    keywords='aiossdb',
    packages=get_packages('aiossdb'),
    extras_require={
        'tests': [
            'pytest',
            'pytest-cov',
            'pytest-asyncio',
        ],
    },
    classifiers=[
        'Development Status :: 3 - Alpha',
        'Programming Language :: Python :: 3',
        'Programming Language :: Python :: 3.6',
        'Intended Audience :: Developers',
        'License :: OSI Approved :: MIT License',
        'Operating System :: OS Independent',
        'Topic :: Software Development :: Libraries :: Python Modules',
    ],
)
